import io

from mallo.request import Request


def make_environ(path='/', method='GET', query_string='', headers=None, body=b''):
    headers = headers or {}
    environ = {
        'REQUEST_METHOD': method,
        'PATH_INFO': path,
        'QUERY_STRING': query_string,
        'SERVER_NAME': 'localhost',
        'SERVER_PORT': '8000',
        'wsgi.version': (1, 0),
        'wsgi.url_scheme': 'http',
        'wsgi.input': io.BytesIO(body),
        'wsgi.errors': io.StringIO(),
        'wsgi.multithread': False,
        'wsgi.multiprocess': False,
        'wsgi.run_once': False,
        'CONTENT_LENGTH': str(len(body)),
    }

    for key, value in headers.items():
        normalized = key.upper().replace('-', '_')
        if normalized in ('CONTENT_TYPE', 'CONTENT_LENGTH'):
            environ[normalized] = value
        else:
            environ[f'HTTP_{normalized}'] = value
    return environ


def test_is_xhr_header():
    env = make_environ(headers={'X-Requested-With': 'XMLHttpRequest'})
    req = Request(env)
    assert req.is_xhr() is True


def test_multipart_form_and_file():
    boundary = '----WebKitFormBoundary7MA4YWxkTrZu0gW'
    body = (
        f'--{boundary}\r\n'
        'Content-Disposition: form-data; name="name"\r\n\r\n'
        'Mallo\r\n'
        f'--{boundary}\r\n'
        'Content-Disposition: form-data; name="file"; filename="hello.txt"\r\n'
        'Content-Type: text/plain\r\n\r\n'
        'Hello file\r\n'
        f'--{boundary}--\r\n'
    ).encode()

    env = make_environ(
        method='POST',
        headers={'Content-Type': f'multipart/form-data; boundary={boundary}'},
        body=body
    )
    req = Request(env)
    assert req.form.get('name') == 'Mallo'
    assert 'file' in req.files
    assert req.files['file']['filename'] == 'hello.txt'
    assert req.files['file']['content'] == b'Hello file'



"""
Utility functions for Mallo framework
"""

import os
import re
import hashlib
import json
from typing import Any, Dict

def secure_filename(filename: str)-> str:
    """
    make a filename secure for filesystem storage
    Args:
        filename: Original filename

    Returns:
         Secure version of filename

    :param filename:
    :return:
    """
    # Remove path separators
    filename = filename.replace('/','_').replace('\\','_')

    #Remove any non-printable characters
    filename = re.sub(r'[^\w\s.-]', '', filename)

    # Limit length
    if len(filename) > 255:
        name, ext = os.path.splitext(filename)
        filename = name[:255-len(ext)] + ext
    return filename.strip()

def generate_etag(content: Any) -> str:
    """
    Generate Etag for content

    Args:
       content: Content to generate Etag for

    Returns:
        ETag string

    :param content:
    :return:
    """
    if isinstance(content, str):
        content = content.encode()
    elif isinstance(content, (dict, list)):
        content = json.dumps(content, sort_keys=True)

    return hashlib.md5(content).hexdigest()

def parse_multipart_form(headers, body):
    """
    Parse multipart form data
    (To be implemented for file uploads)

    :param headers:
    :param body:
    :return:
    """
    content_type = headers.get('Content-Type', '')
    if 'boundary=' not in content_type:
        return {}, {}

    boundary = content_type.split('boundary=')[-1].strip()
    if boundary.startswith('"') and boundary.endswith('"'):
        boundary = boundary[1:-1]

    boundary_bytes = ('--' + boundary).encode()
    parts = body.split(boundary_bytes)

    form = {}
    files = {}

    for part in parts:
        part = part.strip(b'\r\n')
        if not part or part == b'--':
            continue

        if b'\r\n\r\n' not in part:
            continue

        raw_headers, content = part.split(b'\r\n\r\n', 1)
        raw_headers = raw_headers.decode('utf-8', errors='ignore').split('\r\n')
        header_map = {}
        for line in raw_headers:
            if ':' in line:
                key, value = line.split(':', 1)
                header_map[key.strip().lower()] = value.strip()

        disposition = header_map.get('content-disposition', '')
        if 'form-data' not in disposition:
            continue

        name = None
        filename = None
        for item in disposition.split(';'):
            item = item.strip()
            if item.startswith('name='):
                name = item.split('=', 1)[1].strip('"')
            elif item.startswith('filename='):
                filename = item.split('=', 1)[1].strip('"')

        if not name:
            continue

        if content.endswith(b'\r\n'):
            content = content[:-2]

        if filename is not None:
            content_type_value = header_map.get('content-type', 'application/octet-stream')
            files[name] = {
                'filename': filename,
                'content_type': content_type_value,
                'content': content
            }
        else:
            value = content.decode('utf-8', errors='ignore')
            if name in form:
                if isinstance(form[name], list):
                    form[name].append(value)
                else:
                    form[name] = [form[name], value]
            else:
                form[name] = value

    return form, files

def slugify(text: str) -> str:
    """
    Convert text to URL-friendly slug

    Args:
        text: Text to convert

    Return:
        Slug version of text
    :param text:
    :return:
    """
    # Convert to lowercase
    text = text.lower()

    # Replace non-alphanumeric with hyphen
    text = re.sub(r'[^a-z0-9]+', '-', text)

    text = text.strip('-')

    return text

def camel_to_snake(name: str) -> str:
    """
    Convert camelcase to snake_case

    Args:
        name: camelCase string

    Returns:
          snake_case string

    :param name:
    :return:
    """
    pattern = re.compile(r'(?<!^)(?=[A-Z])')
    return pattern.sub('_', name).lower()


def snake_to_camel(name: str) -> str:
    """
    Convert snake_cas to camelCase

    Args:
        name: snake_case string

    Returns:
          camelCase string

    :param name:
    :return:
    """
    components = name.split('_')
    return components[0] + ''.join(x.title() for x in components[1:])



"""
Routing System for Mallo framework
"""
import re
from typing import Dict, List, Tuple, Optional, Callable

class Router:
    """
    Handles URL routing with support for dynamic  parameters

    Dynamic route example:
      /user/<name>           -> matches /user/betrand, passes {'name': 'betrand'}
      /post/<int:id>         -> matches /post/123, passes {'id': 123}
      /file/<path:filepath>  -> matches /file/documents/test.txt

    """

    def __init__(self):
        self.routes = [] # List of (patterns, handler, methods)

    def add_route(self, path: str, handler: Callable, methods: List[str], options: Optional[Dict] = None):
        """
        Add a new route to the router

        Args:
            path: URL path with optional parameters
            handler: Function to handle the route
            methods: List of HTTP methods

        :param path:
        :param handler:
        :param methods:
        """
        # Convert path to regex pattern
        pattern = self._path_to_pattern(path)
        options = options or {}

        for method in methods:
            self.routes.append({
                'pattern': pattern,
                'handler': handler,
                'method': method.upper(),
                'original_path': path,
                'options': options.copy()
            })

    def _path_to_pattern(self, path: str):
        """
        Convert a path with parameter to a regex pattern

        Supports:
                <name>     -> string parameter
                <int: id>  -> integer parameter
                <float: value> -> float parameter
                <path: filepath> -> path parameter

        Example:
              /user/<int:id>/<name> -> /user/123/betrand
        :param path:
        :return:
        """
        param_names = []

        # Replace parameter placeholders
        def replace_param(match):
            param = match.group(1)

            # Check for type specification
            if ':' in param:
                param_type, param_name = param.split(':')
                param_names.append((param_name, param_type))

                # Return appropriate regex pattern
                if param_type == 'int':
                    return r'(\d+)'
                elif param_type == 'float':
                    return r'(\d+\.?\d*)'
                elif param_type == 'path':
                    return r'(.+)'
                else: # string (default)
                    return r'([^/]+)'
            else:
                # Default to string
                param_names.append((param, 'str'))
                return r'([^/]+)'

        #replace <....> with regex patterns
        pattern = re.sub(r'<([^>]+)>', replace_param, path)

        #compile the regex
        regex = re.compile(f'^{pattern}$')

        return  regex, param_names


    def match(self, path: str, method: str, return_route: bool = False) -> Optional[Tuple[Callable,Dict]]:
        """
        Match a path and method to a route

        Returns:
              Tuple of (handler, kwargs) or None if no match

        :param path:
        :param method:
        :return:
        """
        for route in self.routes:
            if route['method'] != method.upper():
                continue

            pattern, param_types = route['pattern']
            match = pattern.match(path)

            if match:
                # Build kwargs from captured groups
                kwargs = {}
                for i, (name, param_type) in enumerate(param_types):
                    value = match.group(i + 1)

                    # Convert to appropriate type
                    if param_type == 'int':
                        value = int(value)
                    elif param_type == 'float':
                        value = float(value)
                    # 'str' and 'path' remain as string

                    kwargs[name] = value

                if return_route:
                    return route['handler'], kwargs, route
                return route['handler'], kwargs

        return None

    def url_for(self, handler_name: str, **kwargs) -> str:
        """
        Generate URL for a named route (to be implemented)
        :param handler_name:
        :param kwargs:
        :return:
        """
        for route in self.routes:
            handler = route['handler']
            route_name = route.get('options', {}).get('name')
            resolved_name = route_name or getattr(handler, '__name__', None)
            if resolved_name != handler_name:
                continue

            path = route['original_path']

            def replace_param(match):
                param = match.group(1)
                if ':' in param:
                    _, param_name = param.split(':', 1)
                else:
                    param_name = param

                if param_name not in kwargs:
                    raise KeyError(f"Missing route param: {param_name}")
                return str(kwargs[param_name])

            return re.sub(r'<([^>]+)>', replace_param, path)

        raise KeyError(f"No route found for handler: {handler_name}")



"""
Configuration utilities for Mallo.
"""

import os


class MalloConfig:
    """
    App config with environment variable overrides.
    Explicit constructor values take precedence over environment values.
    """

    DEFAULTS = {
        'debug': False,
        'live_reload': True,
        'static_url_path': '/static',
        'static_folder': 'static',
        'template_folder': 'templates',
        'auto_escape': True,
        'enable_logging': True,
        'security_headers': True,
        'static_cache_seconds': None,
        'secret_key': None,
        'csrf_protect': True,
        'session_cookie': 'mallo_session',
        'error_page_404': None,
        'error_page_500': None,
    }

    ENV_MAP = {
        'debug': ('MALLO_DEBUG', 'bool'),
        'live_reload': ('MALLO_LIVE_RELOAD', 'bool'),
        'static_url_path': ('MALLO_STATIC_URL_PATH', 'str'),
        'static_folder': ('MALLO_STATIC_FOLDER', 'str'),
        'template_folder': ('MALLO_TEMPLATE_FOLDER', 'str'),
        'auto_escape': ('MALLO_AUTO_ESCAPE', 'bool'),
        'enable_logging': ('MALLO_ENABLE_LOGGING', 'bool'),
        'security_headers': ('MALLO_SECURITY_HEADERS', 'bool'),
        'static_cache_seconds': ('MALLO_STATIC_CACHE_SECONDS', 'int_or_none'),
        'secret_key': ('MALLO_SECRET_KEY', 'str'),
        'csrf_protect': ('MALLO_CSRF_PROTECT', 'bool'),
        'session_cookie': ('MALLO_SESSION_COOKIE', 'str'),
        'error_page_404': ('MALLO_ERROR_PAGE_404', 'str'),
        'error_page_500': ('MALLO_ERROR_PAGE_500', 'str'),
    }

    def __init__(self, **overrides):
        self._data = dict(self.DEFAULTS)

        for key, (env_key, cast_type) in self.ENV_MAP.items():
            raw = os.environ.get(env_key)
            if raw is None:
                continue
            self._data[key] = self._cast(raw, cast_type)

        for key, value in overrides.items():
            self._data[key] = value

        if self._data['static_cache_seconds'] is None:
            self._data['static_cache_seconds'] = 0 if self._data['debug'] else 3600

    def _cast(self, value, cast_type):
        if cast_type == 'str':
            return value
        if cast_type == 'bool':
            return str(value).strip().lower() in ('1', 'true', 'yes', 'on')
        if cast_type == 'int_or_none':
            text = str(value).strip().lower()
            if text in ('none', 'null', ''):
                return None
            return int(text)
        return value

    def get(self, key, default=None):
        return self._data.get(key, default)

    def set(self, key, value):
        self._data[key] = value

    def as_dict(self):
        return dict(self._data)



import io

from mallo import Mallo


def make_environ(path='/', method='GET', headers=None, body=b''):
    headers = headers or {}
    environ = {
        'REQUEST_METHOD': method,
        'PATH_INFO': path,
        'QUERY_STRING': '',
        'SERVER_NAME': 'localhost',
        'SERVER_PORT': '8000',
        'wsgi.version': (1, 0),
        'wsgi.url_scheme': 'http',
        'wsgi.input': io.BytesIO(body),
        'wsgi.errors': io.StringIO(),
        'wsgi.multithread': False,
        'wsgi.multiprocess': False,
        'wsgi.run_once': False,
        'CONTENT_LENGTH': str(len(body)),
    }

    for key, value in headers.items():
        normalized = key.upper().replace('-', '_')
        if normalized in ('CONTENT_TYPE', 'CONTENT_LENGTH'):
            environ[normalized] = value
        else:
            environ[f'HTTP_{normalized}'] = value
    return environ


def call_app(app, environ):
    captured = {}

    def start_response(status, headers):
        captured['status'] = status
        captured['headers'] = dict(headers)

    body = b''.join(app(environ, start_response))
    return captured['status'], captured['headers'], body


def test_sessions_and_csrf():
    app = Mallo(__name__, secret_key='test-secret')

    @app.get('/form')
    def form(request):
        return 'ok'

    @app.post('/submit')
    def submit(request):
        return 'done'

    status, headers, body = call_app(app, make_environ('/form', 'GET'))
    assert status.startswith('200')
    assert 'Set-Cookie' in headers
    cookie_value = headers['Set-Cookie'].split(';', 1)[0].split('=', 1)[1].strip()
    session_id = cookie_value.split('|', 1)[0]
    csrf_token = app._session_store[session_id]['csrf_token']

    post_body = f'csrf_token={csrf_token}'.encode()
    status, headers, body = call_app(app, make_environ(
        '/submit',
        'POST',
        headers={
            'Content-Type': 'application/x-www-form-urlencoded',
            'Cookie': f'mallo_session={cookie_value}',
        },
        body=post_body
    ))
    assert status.startswith('200')

    status, headers, body = call_app(app, make_environ(
        '/submit',
        'POST',
        headers={
            'Content-Type': 'application/x-www-form-urlencoded',
            'Cookie': f'mallo_session={cookie_value}',
        },
        body=b''
    ))
    assert status.startswith('403')


def test_route_level_csrf_override():
    app = Mallo(__name__, secret_key='test-secret', csrf_protect=True)

    @app.post('/webhook', csrf=False)
    def webhook(request):
        return 'ok'

    status, headers, body = call_app(app, make_environ(
        '/webhook',
        'POST',
        headers={'Content-Type': 'application/x-www-form-urlencoded'},
        body=b''
    ))
    assert status.startswith('200')



"""
Response object for mallo Framework
"""

import json
from typing import Any, Dict, Optional

class Response:
    """
    Represents an HTTP response

    Example:
        return Response('hello world', status=200, header={'Content-Type': 'text/plain'})
    """
    def __init__(self, body='',status=200,headers=None,content_type='text/html'):
        self.body = body
        self.status = status
        self.headers = headers or {}

        # Set a content type if not already set
        if 'Content-Type' not in self.headers:
            if isinstance(body, (dict, list)):
                self.headers['Content-Type'] = 'application/json'
            else:
                self.headers['Content-Type'] =content_type

        #add default headers
        self.headers['Server'] = 'Mallo'

    def to_wsgi(self, start_response):
        """
        Convert to WSGI response

        :param start_response:
        :return:
        """
        status_map = {
            200: '200 ok',
            201: '201 Created',
            204: '204 No Content',
            301: '301 Moved Permanently',
            302: '302 Found',
            304: '304 Not Modified',
            400: '400 Bad Request',
            401: '401 Unauthorised',
            403: '403 Forbidden',
            404: '404 Not Found',
            405: '405 Method Not Allowed',
            418: "418 I'm a teapot",
            500: '500 Internal Server Error',
            502: '502 Bad Gateway',
            503: '503 Service Unavailable',
        }
        status_text = status_map.get(self.status, f'{self.status} Unknown')

        # Convert header to List of tuples
        headers_list = [(k,v) for k, v in self.headers.items()]

        start_response(status_text, headers_list)

        #Encode body if it is a string
        if isinstance(self.body, str):
            return [self.body.encode()]
        elif isinstance(self.body, (dict, list)):
            # Auto-convert dict/list to json
            return [json.dumps(self.body).encode()]
        else:
            return [self.body]

    def set_cookie(self, key, value='', max_age=None, expires=None, path='/', domain=None, secure=False, httponly=False):
        """
        Set a cookie in the response

        :param key:
        :param value:
        :param max_age:
        :param expires:
        :param path:
        :param domain:
        :param secure:
        :param httponly:
        :return:
        """
        cookie = f'{key} = {value}'
        if max_age is not None:
            cookie += f'; Max-Age={max_age}'
        if expires is not None:
            cookie += f'; Expires={expires}'
        if path is not None:
            cookie += f'; Path={path}'
        if domain is not None:
            cookie += f'; Domain={domain}'
        if secure:
            cookie += '; Secure'
        if httponly:
            cookie += '; HttpOnly'

        self.headers['Set-Cookie'] = cookie

    def delete_cookie(self, key, path='/', domain=None):
        """
        Delete a cookie by setting it to expire immediately

        :param key:
        :param path:
        :param domain:
        :return:
        """
        self.set_cookie(key, '',expires='Thu, 01 Jan 1970 00:00:00 GMT', path=path, domain=domain)



class JSONResponse(Response):
    """
    Response with JSON content

    """
    def __init__(self, data, status=200, headers=None):
        super().__init__(
            body=json.dumps(data),
            status=status,
            headers=headers,
            content_type='application/json'
        )


class RedirectResponse(Response):
    """
    Redirect response
    """
    def __init__(self, location, status=302):
        super().__init__(
            body='',
            status=status,
            headers={'Location': location}
        )

class FileResponse(Response):
    """
    File download response
    """
    def __init__(self, filepath, filename=None, status=200, headers=None):
        with open(filepath, 'rb') as f:
            content = f.read()

        if filename is None:
            import os
            filename = os.path.basename(filepath)

        headers = headers or {}
        headers['Content-Disposition'] = f'attachment; filename="{filename}"'

        super().__init__(
            body=content,
            status=status,
            headers=headers,
            content_type='application/octet-stream'
        )



"""
Request object for Mallo framework
"""

import json
from urllib.parse import parse_qs
from typing import Dict, Any, Optional
from mallo.utils import parse_multipart_form

class Request:
    """
    Represents an HTTP request

    Attributes:
        method: HTTP method (GET, POST, etc.)
        path: Request path
        headers: Dictionary of HTTP headers
        query: Dictionary of query parameter
        form: Dictionary of form data (fo post requests)
        json: parsed JSON body (if applicable)
        body: Raw request body
        cookies: Dictionary of cookies
    """

    def __init__(self, environ):
        self.eviron = environ
        self.method = environ['REQUEST_METHOD'].upper()
        self.path = environ['PATH_INFO']
        self.query_string = environ.get('QUERY_STRING', '')

        # Parse headers
        self.headers = self._parse_headers(environ)

        # Parse query parameter
        self.query = parse_qs(self.query_string)
        # Flatten single-value query params
        self.query = {k: v[0] if len(v) == 1 else v for k, v in self.query.items()}

        #parse body
        self.body = self._get_body()

        # Parse form data
        self.form = {}
        self.files = {}
        self.json = None

        content_type = self.headers.get('Content-Type', '')
        if self.method == 'POST' and self.body:
            if 'application/x-www-form-urlencoded' in content_type:
                self.form = parse_qs(self.body.decode())
                # Flatten single-value form fields
                self.form = {k: v[0] if len(v) == 1 else v for k, v in self.form.items()}
            elif 'application/json' in content_type:
                try:
                    self.json = json.loads(self.body.decode())
                except:
                    self.json = None
            elif 'multipart/form-data' in content_type:
                try:
                    form, files = parse_multipart_form(self.headers, self.body)
                    self.form = form
                    self.files = files
                except:
                    self.form = {}
                    self.files = {}

        # Parse cookies
        self.cookies = self._parse_cookies()

    def _parse_headers(self, environ):
        """
        Parse HTTP headers from WSGI environ

        :param environ:
        :return:
        """
        headers = {}
        for key, value in environ.items():
            if key.startswith('HTTP_'):
                header_name = key[5:].replace('_','-').title()
                headers[header_name] = value
            elif key in ('CONTENT_TYPE', 'CONTENT_LENGTH'):
                headers[key.replace('_','-').title()] = value
        return headers

    def _get_body(self):
        """
        Read request body

        :return:
        """
        try:
            content_length = int(self.eviron.get('CONTENT_LENGTH', 0))
            if content_length > 0:
                return self.eviron['wsgi.input'].read(content_length)
        except:
            pass
        return b''

    def _parse_cookies(self):
        """
        Parse cookies from header

        :return:
        """
        cookies = {}
        cookie_header = self.headers.get('Cookie', '')

        if cookie_header:
            for cookie in cookie_header.split(';'):
                if '=' in cookie:
                    name, value = cookie.strip().split('=', 1)
                    cookies[name] = value

        return cookies

    def get(self, key, default=None):
        """
        Get a query parameter

        :param key:
        :param default:
        :return:
        """
        return self.query.get(key, default)

    def post(self, key, default=None):
        """
        Get a form parameter

        :param key:
        :param default:
        :return:
        """
        return self.form.get(key, default)

    def is_xhr(self):
        """
        Check if request is AJAX

        :return:
        """
        return self.headers.get('X-Requested-With') == 'XMLHttpRequest'

    def __repr__(self):
        return f'<Request {self.method} {self.path}>'



"""
Database support for Mallo using SQLAlchemy Core.
"""

from contextlib import contextmanager

try:
    from sqlalchemy import create_engine, text
except Exception:  # pragma: no cover - handled at runtime
    create_engine = None
    text = None


class Database:
    """
    Minimal SQLAlchemy Core database wrapper.
    """

    def __init__(self, url: str, **engine_kwargs):
        if create_engine is None:
            raise RuntimeError(
                "SQLAlchemy is required for database support. "
                "Install it with: pip install sqlalchemy"
            )
        self.url = url
        self.engine = create_engine(url, future=True, **engine_kwargs)

    def execute(self, sql: str, params=None):
        """
        Execute write or generic SQL in a transaction.
        """
        params = params or {}
        with self.engine.begin() as conn:
            return conn.execute(text(sql), params)

    def fetchone(self, sql: str, params=None):
        """
        Execute a SELECT and return one row as dict (or None).
        """
        params = params or {}
        with self.engine.connect() as conn:
            result = conn.execute(text(sql), params)
            row = result.mappings().first()
            return dict(row) if row else None

    def fetchall(self, sql: str, params=None):
        """
        Execute a SELECT and return all rows as dict list.
        """
        params = params or {}
        with self.engine.connect() as conn:
            result = conn.execute(text(sql), params)
            return [dict(row) for row in result.mappings().all()]

    @contextmanager
    def transaction(self):
        """
        Transaction context manager.
        """
        with self.engine.begin() as conn:
            yield conn

    def close(self):
        """
        Dispose underlying SQLAlchemy engine.
        """
        self.engine.dispose()



"""
Hot reload functionality for Mallo Framework
"""
import os
import sys
import time
import signal
import subprocess
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

class ReloadHandler(FileSystemEventHandler):
    """
    Handles file system events and triggers reload
    """
    def __init__(self, callback, watch_paths=None):
        self.callback = callback
        self.last_trigger = 0
        self.cooldown = 1.0 # Cooldown period in seconds
        self.watch_extensions = {
            '.py', '.html', '.htm', '.css', '.js', '.json', '.txt'
        }
        self.ignore_parts = {
            '__pycache__', '.pytest_cache', '.git', '.venv', '.idea'
        }
        self.file_signatures = {}
        self._prime_signatures(watch_paths or ['.'])

    def _signature(self, path):
        try:
            stat = os.stat(path)
            return (stat.st_mtime_ns, stat.st_size)
        except OSError:
            return None

    def _prime_signatures(self, watch_paths):
        for root in watch_paths:
            if not os.path.exists(root):
                continue
            for dirpath, dirnames, filenames in os.walk(root):
                normalized_dir = dirpath.replace('\\', '/')
                if any(part in normalized_dir for part in self.ignore_parts):
                    continue

                for filename in filenames:
                    path = os.path.join(dirpath, filename)
                    if not self._should_reload(path):
                        continue
                    sig = self._signature(path)
                    if sig is not None:
                        self.file_signatures[path] = sig

    def _should_reload(self, src_path):
        normalized = src_path.replace('\\', '/')

        # Ignore noisy/system paths.
        if any(part in normalized for part in self.ignore_parts):
            return False

        filename = normalized.rsplit('/', 1)[-1]

        # Ignore common temporary editor files.
        if (
            filename.startswith('.')
            or filename.endswith('~')
            or filename.endswith('.tmp')
            or filename.endswith('.swp')
            or filename.endswith('.swx')
            or filename.endswith('.bak')
        ):
            return False

        _, ext = os.path.splitext(filename.lower())
        return ext in self.watch_extensions

    def _process_event_path(self, src_path):
        if not self._should_reload(src_path):
            return

        current_signature = self._signature(src_path)
        previous_signature = self.file_signatures.get(src_path)

        if current_signature is None:
            return

        if previous_signature is None:
            # New file: track and trigger reload once.
            self.file_signatures[src_path] = current_signature
        elif previous_signature == current_signature:
            # Duplicate/no-op event.
            return
        else:
            self.file_signatures[src_path] = current_signature

        current_time = time.time()
        if current_time - self.last_trigger > self.cooldown:
            self.last_trigger = current_time
            print(f"\n * Detected change in: {src_path}")
            self.callback()

    def on_modified(self, event):
        # Ignore directory modifications
        if event.is_directory:
            return
        self._process_event_path(event.src_path)

    # Some editors save via "create new file + move/replace".
    def on_created(self, event):
        if event.is_directory:
            return
        self._process_event_path(event.src_path)

    def on_moved(self, event):
        if event.is_directory:
            return
        self._process_event_path(event.dest_path)

class HotReloader:
    """
    Manages hot reload functionality
    """
    def __init__(self, app):
        self.app = app
        self.process = None
        self.observer = None
        self.watch_paths = ['.'] # Directory to watch

    def run(self, host, port):
        """
        Start the server with hot reload
        :param host:
        :param port:
        :return:
        """
        print("* Hot reload is enable")
        print("* Watching for file changes...")

        #Start server in a subprocess
        self.start_server_process(host, port)

        # Start file watcher
        self.start_watcher()

        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            self.stop()

    def start_server_process(self, host, port):
        """
        Start the server as a subprocess

        :param host:
        :param port:
        :return:
        """
        env = os.environ.copy()
        env['MALLO_HOT_RELOAD'] = '1'

        # Get the main script path
        main_script = sys.argv[0]

        self.process = subprocess.Popen(
            [sys.executable, main_script],
            env=env
        )

    def start_watcher(self):
        """
        Start watching for file changes
        :return:
        """
        self.observer = Observer()
        handler = ReloadHandler(self.reload_server, self.watch_paths)

        # Watch python files in current directory and subdirectories
        for path in self.watch_paths:
            if os.path.exists(path):
                self.observer.schedule(handler, path, recursive=True)

        self.observer.start()

    def reload_server(self):
        """
        Reload the server process

        :return:
        """
        print(" * Reloading server....")

        # kill the old process
        if self.process:
            self.process.terminate()
            try:
                self.process.wait(timeout=5)
            except subprocess.TimeoutExpired:
                self.process.kill()

        # get the main script path and arguments
        main_script = sys.argv[0]

        # start new process
        env = os.environ.copy()
        env['MALLO_HOT_RELOAD'] = '1'

        self.process = subprocess.Popen(
            [sys.executable, main_script] + sys.argv[1:],
            env=env
        )

        print("* Reload complete!")

    def stop(self):
        """
        Stop the hot reloader and server

        :return:
        """
        print("\n * Stopping server....")
        if self.observer:
            self.observer.stop()
            self.observer.join()

        if self.process:
            self.process.terminate()
            try:
                self.process.wait(timeout=5)
            except subprocess.TimeoutExpired:
                self.process.kill()

        print("* Server stopped")



from mallo.template import TemplateEngine


def test_template_if_and_for():
    engine = TemplateEngine()
    template = "Hello {% if show %}Yes{% endif %} {% for item in items %}{{ item }} {% endfor %}"
    result = engine._render_string(template, {'show': True, 'items': ['A', 'B']})
    assert "Hello Yes" in result
    assert "A" in result
    assert "B" in result



"""
Tests for Mallo application
"""

import io
import pytest
from mallo import Mallo, Request, Response


def test_app_creation():
    """Test basic app creation"""
    app = Mallo(__name__)
    assert app.import_name == __name__
    assert app.router is not None


def test_route_decorator():
    """Test route registration"""
    app = Mallo(__name__)

    @app.route('/')
    def home(request):
        return 'Home'

    matching = [route for route in app.router.routes if route['original_path'] == '/']
    assert len(matching) == 1
    assert matching[0]['method'] == 'GET'


def test_dynamic_route():
    """Test dynamic route matching"""
    app = Mallo(__name__)

    @app.route('/user/<int:id>')
    def user(request, id):
        return f'User {id}'

    match = app.router.match('/user/123', 'GET')
    assert match is not None
    handler, kwargs = match
    assert kwargs == {'id': 123}


def test_response_json_content_type():
    response = Response({'a': 1})
    captured = {}

    def start_response(status, headers):
        captured['status'] = status
        captured['headers'] = dict(headers)

    body = b''.join(response.to_wsgi(start_response))
    assert captured['headers'].get('Content-Type') == 'application/json'
    assert body == b'{"a": 1}'


def test_middleware_chain():
    app = Mallo(__name__)

    @app.middleware
    def add_header(request, call_next):
        response = call_next(request)
        if isinstance(response, Response):
            response.headers['X-Middleware'] = 'ok'
        return response

    @app.get('/hello')
    def hello(request):
        return 'hello'

    captured = {}

    def start_response(status, headers):
        captured['status'] = status
        captured['headers'] = dict(headers)

    environ = {
        'REQUEST_METHOD': 'GET',
        'PATH_INFO': '/hello',
        'QUERY_STRING': '',
        'SERVER_NAME': 'localhost',
        'SERVER_PORT': '8000',
        'wsgi.version': (1, 0),
        'wsgi.url_scheme': 'http',
        'wsgi.input': io.BytesIO(b''),
        'wsgi.errors': io.StringIO(),
        'wsgi.multithread': False,
        'wsgi.multiprocess': False,
        'wsgi.run_once': False,
        'CONTENT_LENGTH': '0',
    }

    body = b''.join(app(environ, start_response))
    assert captured['status'].startswith('200')
    assert captured['headers'].get('X-Middleware') == 'ok'
    assert body == b'hello'


def test_route_group_prefix_and_url_name():
    app = Mallo(__name__)
    api = app.group('/api', middleware=[])

    @api.get('/users/<int:id>', name='api_user')
    def user(request, id):
        return f'user-{id}'

    assert app.url_for('api_user', id=7) == '/api/users/7'


def test_env_override_template_folder(monkeypatch):
    monkeypatch.setenv('MALLO_TEMPLATE_FOLDER', 'views')
    app = Mallo(__name__)
    assert app.config_obj.get('template_folder') == 'views'



"""
Core Application for Mallo framework
"""

import os
import sys
import time
import hmac
import hashlib
import secrets
import logging
import html
from email.utils import formatdate
from wsgiref.simple_server import make_server, WSGIRequestHandler
from mallo.router import Router
from mallo.request import Request
from mallo.response import Response
from mallo.template import render_template, render_template_file
from mallo.config import MalloConfig
from mallo.utils import generate_etag
from mallo.hot_reload import HotReloader


class RouteGroup:
    """
    Route prefix group with optional default route options.
    """

    def __init__(self, app, prefix, middleware=None, route_defaults=None):
        self.app = app
        self.prefix = prefix.rstrip('/') or '/'
        self.middleware = middleware or []
        self.route_defaults = route_defaults or {}

    def _join(self, path):
        if not path.startswith('/'):
            path = '/' + path
        if self.prefix == '/':
            return path
        return self.prefix + path

    def route(self, path, methods=None, **options):
        merged = dict(self.route_defaults)
        merged.update(options)
        route_middleware = []
        route_middleware.extend(self.middleware)
        route_middleware.extend(merged.get('middleware') or [])
        merged['middleware'] = route_middleware
        return self.app.route(self._join(path), methods=methods, **merged)

    def get(self, path, **options):
        return self.route(path, methods=['GET'], **options)

    def post(self, path, **options):
        return self.route(path, methods=['POST'], **options)

    def put(self, path, **options):
        return self.route(path, methods=['PUT'], **options)

    def delete(self, path, **options):
        return self.route(path, methods=['DELETE'], **options)


class Mallo:
    """
    Main application class for Mallo framework
    Example:
        app = Mallo(__name__)
        @app.route("/")
        def home(request):
            return "<h1>Hello, Mallo!</h1>"

    """

    def __init__(self, import_name, static_url_path="/static", **config):
        self.import_name = import_name
        merged_config = dict(config)
        merged_config['static_url_path'] = merged_config.get('static_url_path', static_url_path)
        self.config_obj = MalloConfig(**merged_config)
        self.config = self.config_obj.as_dict()
        self.static_url_path = self.config_obj.get('static_url_path')
        self.router = Router()
        self.hot_reloader = None
        self.debug = self.config_obj.get('debug')
        self.live_reload = self.config_obj.get('live_reload')
        self._reload_token = f"{os.getpid()}-{time.time_ns()}"
        self.secret_key = self.config_obj.get('secret_key')
        self.csrf_protect = self.config_obj.get('csrf_protect')
        self._session_cookie = self.config_obj.get('session_cookie')
        self._session_store = {}
        self.before_request_funcs = []
        self.after_request_funcs = []
        self.middlewares = []
        self._db_bindings = []
        self.error_handlers = {}
        self._enable_logging = self.config_obj.get('enable_logging')
        self._security_headers = self.config_obj.get('security_headers')
        self._static_cache_seconds = self.config_obj.get('static_cache_seconds')
        self.logger = logging.getLogger('mallo')
        self._default_404 = os.path.join(os.path.dirname(__file__), 'defaults', '404.html')
        self._default_500 = os.path.join(os.path.dirname(__file__), 'defaults', '500.html')

        # Add static route if static folder exists
        self._setup_static_routing()


    def _setup_static_routing(self):
        """ Setup static file serving if static folder exists """
        static_folder = self.config_obj.get('static_folder')
        if os.path.exists(static_folder):
            @self.route(f'{self.static_url_path}/<path:filename>')
            def serve_static(request, filename):
                filepath = os.path.join(static_folder, filename)
                if os.path.exists(filepath) and os.path.isfile(filepath):
                    with open(filepath, 'rb') as f:
                        content = f.read()
                    etag = generate_etag(content)
                    last_modified = formatdate(os.path.getmtime(filepath), usegmt=True)
                    if request.headers.get('If-None-Match') == etag:
                        response = Response('', status=304)
                        response.headers['ETag'] = etag
                        response.headers['Last-Modified'] = last_modified
                        return response
                    response = Response(content, headers={
                        'Content-Type': self._guess_mime_type(filename)
                    })
                    response.headers['ETag'] = etag
                    response.headers['Last-Modified'] = last_modified
                    if self._static_cache_seconds > 0:
                        response.headers['Cache-Control'] = f'public, max-age={self._static_cache_seconds}'
                    else:
                        response.headers['Cache-Control'] = 'no-store'
                    return response
                return Response('File not found', status = 404)

    def route(self, path, methods=None, name=None, middleware=None, csrf=None):
        """
        Decorator to register a route

        Args:
            path: URL path (e.g., '/', '/user/<name>')
            methods: List of HTTP methods (default: ['GET'])

        Example:
            @app.route('/hello/<name>')
            def hello(request, name):
                return f'Hello, {name}!'

        :param path
        :param methods
        :return:
        """
        if methods is None:
            methods = ['GET']

        def decorator(handler):
            route_name = name or getattr(handler, '__name__', None)
            options = {
                'name': route_name,
                'middleware': middleware or [],
                'csrf': csrf
            }
            self.router.add_route(path, handler, methods, options=options)
            return handler
        return decorator

    def get(self, path, **options):
        """
        Shorthand for the route with GET method
        :param path:
        :return:
        """
        return self.route(path, methods=['GET'], **options)

    def post(self, path, **options):
        """
        Shorthand for route with POST method
        :param path:
        :return:
        """
        return self.route(path, methods=['POST'], **options)

    def put(self, path, **options):
        """
        Shorthand for route with PUT method
        :param path:
        :return:
        """
        return self.route(path, methods=['PUT'], **options)

    def delete(self, path, **options):
        """
        Shorthand for route with DELETE method
        :param path:
        :return:
        """
        return self.route(path, methods=['DELETE'], **options)

    def group(self, prefix, middleware=None, **route_defaults):
        """
        Create a route group with a shared URL prefix and default options.
        """
        return RouteGroup(self, prefix, middleware=middleware, route_defaults=route_defaults)

    def render_template(self, template_name, **context):
        """
        Render a template from any file path

        Args:
            template_name: template file name relative to template folder
            **context: Variable to pass to template

        Returns:
           Rendered template as a string

        Example:
          @app.route('/')
          def home(request):
              return app.render_template('home.html', name='Mallo User')


        :param template_name:
        :param context:
        :return:
        """
        auto_escape = self.config_obj.get('auto_escape')
        template_folder = self.config_obj.get('template_folder')
        return render_template(
            template_name,
            template_folder=template_folder,
            auto_reload=self.debug,
            auto_escape=auto_escape,
            **context
        )

    def before_request(self, func):
        self.before_request_funcs.append(func)
        return func

    def after_request(self, func):
        self.after_request_funcs.append(func)
        return func

    def init_db(self, database, request_attr='db'):
        """
        Bind a Database object to each request.
        Example:
            db = Database("sqlite:///app.db")
            app.init_db(db)
            # then inside handlers: request.db.fetchall(...)
        """
        self._db_bindings.append((request_attr, database))

        @self.before_request
        def _inject_db(request):
            setattr(request, request_attr, database)

    def middleware(self, func):
        """
        Register a middleware function.
        Signature: middleware(request, call_next) -> Response|str|dict|list
        """
        self.middlewares.append(func)
        return func

    def use(self, middleware_func):
        """
        Register middleware without decorator syntax.
        """
        self.middlewares.append(middleware_func)
        return middleware_func

    def errorhandler(self, status_code: int):
        def decorator(handler):
            self.error_handlers[status_code] = handler
            return handler
        return decorator

    def url_for(self, handler_name: str, **kwargs):
        return self.router.url_for(handler_name, **kwargs)

    def _sign_session_id(self, session_id: str) -> str:
        secret = (self.secret_key or '').encode()
        return hmac.new(secret, session_id.encode(), hashlib.sha256).hexdigest()

    def _make_session_cookie(self, session_id: str) -> str:
        return f"{session_id}|{self._sign_session_id(session_id)}"

    def _load_session(self, request):
        if not self.secret_key:
            request.session = None
            request.session_id = None
            request._session_dirty = False
            return

        raw = request.cookies.get(self._session_cookie, '')
        session_id = None
        if raw and '|' in raw:
            sid, sig = raw.split('|', 1)
            if hmac.compare_digest(sig, self._sign_session_id(sid)):
                session_id = sid

        if not session_id:
            session_id = secrets.token_urlsafe(24)

        if session_id not in self._session_store:
            self._session_store[session_id] = {}

        class SessionData(dict):
            def __init__(self, parent, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self._parent = parent
            def __setitem__(self, key, value):
                self._parent._session_dirty = True
                return super().__setitem__(key, value)
            def __delitem__(self, key):
                self._parent._session_dirty = True
                return super().__delitem__(key)

        request.session_id = session_id
        request.session = SessionData(request)
        request.session.update(self._session_store[session_id])
        request._session_dirty = False

        if 'csrf_token' not in request.session:
            request.session['csrf_token'] = secrets.token_urlsafe(32)
        request.csrf_token = request.session.get('csrf_token')

    def _ensure_session_cookie(self, request, response):
        if not self.secret_key or request.session_id is None:
            return
        if request._session_dirty:
            self._session_store[request.session_id] = dict(request.session)
            cookie_value = self._make_session_cookie(request.session_id)
            response.set_cookie(self._session_cookie, cookie_value, path='/')

    def _csrf_check(self, request, csrf_required=True):
        if not csrf_required:
            return True
        if not self.secret_key or not self.csrf_protect:
            return True
        if request.method in ('GET', 'HEAD', 'OPTIONS'):
            return True

        token = None
        if request.headers.get('X-Csrf-Token'):
            token = request.headers.get('X-Csrf-Token')
        elif request.form:
            token = request.form.get('csrf_token')
        elif request.json:
            token = request.json.get('csrf_token')

        return token and token == request.csrf_token

    def _apply_security_headers(self, response):
        if not self._security_headers:
            return response
        response.headers.setdefault('X-Content-Type-Options', 'nosniff')
        response.headers.setdefault('X-Frame-Options', 'DENY')
        response.headers.setdefault('Referrer-Policy', 'same-origin')
        return response

    def _handle_error(self, request, status_code, default_path):
        handler = self.error_handlers.get(status_code)
        if handler:
            result = handler(request)
            if not isinstance(result, Response):
                result = Response(result, status=status_code)
            return result

        custom_path = self.config_obj.get(f'error_page_{status_code}')
        template_path = custom_path or default_path
        content = render_template_file(template_path, auto_reload=self.debug, auto_escape=True)
        return Response(content, status=status_code)

    def _user_routes_count(self):
        """
        Count application routes excluding static and internal reload endpoints.
        """
        static_prefix = f'{self.static_url_path}/'
        count = 0
        for route in self.router.routes:
            original = route.get('original_path', '')
            if original.startswith(static_prefix):
                continue
            if original == '/__mallo_reload__':
                continue
            count += 1
        return count

    def _print_startup_diagnostics(self):
        """
        Show concise startup diagnostics for common routing mistakes.
        """
        if self._user_routes_count() > 0:
            return
        print(" ! Warning: No application routes registered.")
        print("   Common causes:")
        print("   - Missing '@' before app.route/app.get/app.post")
        print("   - Route module not imported before app.run()")
        print("   - Route decorator used but function not executed")

    def _friendly_debug_error_response(self, exc):
        """
        Render a clear and non-overwhelming debug error page.
        """
        import traceback

        exc_type = html.escape(type(exc).__name__)
        exc_msg = html.escape(str(exc) or "No message")
        tb = html.escape(traceback.format_exc())

        body = f"""
<html>
  <head>
    <meta charset="utf-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    <title>Mallo Debug Error</title>
    <style>
      body {{ font-family: Segoe UI, Arial, sans-serif; margin: 0; background: #f6f3ee; color: #1f1a15; }}
      .wrap {{ max-width: 780px; margin: 30px auto; padding: 0 16px; }}
      .card {{ background: #fffdf9; border: 1px solid #e8dece; border-radius: 14px; padding: 18px; }}
      .tag {{ display: inline-block; background: #f8e8de; color: #b6521a; border-radius: 999px; padding: 4px 10px; font-size: 12px; font-weight: 700; letter-spacing: .05em; text-transform: uppercase; }}
      h1 {{ margin: 10px 0 8px; font-size: 26px; }}
      p {{ margin: 0 0 12px; color: #5f5549; }}
      code {{ background: #f4efe7; padding: 2px 6px; border-radius: 6px; }}
      details {{ margin-top: 12px; }}
      pre {{ white-space: pre-wrap; word-break: break-word; background: #1a1f29; color: #eaf0ff; border-radius: 10px; padding: 12px; overflow: auto; }}
    </style>
  </head>
  <body>
    <div class="wrap">
      <div class="card">
        <span class="tag">Debug Error</span>
        <h1>{exc_type}</h1>
        <p>{exc_msg}</p>
        <p>Tip: fix the issue and refresh the page.</p>
        <details>
          <summary>Show traceback details</summary>
          <pre>{tb}</pre>
        </details>
      </div>
    </div>
  </body>
</html>
"""
        return Response(body, status=500)

    def _ensure_response(self, result):
        if not isinstance(result, Response):
            return Response(result)
        return result

    def _process_outgoing(self, request, response, start_time):
        """
        Apply common response processing once.
        """
        if getattr(response, "_mallo_processed", False):
            return response

        response = self._apply_security_headers(response)
        self._ensure_session_cookie(request, response)
        if self._enable_logging:
            elapsed = (time.perf_counter() - start_time) * 1000
            self.logger.info('%s %s -> %s (%.2fms)', request.method, request.path, response.status, elapsed)
        response._mallo_processed = True
        return response

    def _run_middleware_chain(self, request, endpoint, middleware_list):
        """
        Run middleware chain and return the final result.
        """
        index = 0

        def call_next(req):
            nonlocal index
            if index >= len(middleware_list):
                return endpoint(req)
            current = middleware_list[index]
            index += 1
            return current(req, call_next)

        return call_next(request)

    def _dispatch_request(self, request, start_time, route_match=None):
        """
        Resolve route and produce a response.
        """
        if route_match is None:
            route_match = self.router.match(request.path, request.method, return_route=True)
        if route_match:
            handler, kwargs, route = route_match
            route_options = route.get('options', {})
            route_middleware = route_options.get('middleware') or []

            def endpoint(req):
                return handler(req, **kwargs)
            try:
                result = self._run_middleware_chain(request, endpoint, route_middleware)
                result = self._ensure_response(result)

                if self.debug:
                    self._set_no_cache_headers(result)
                if self.debug and self.live_reload:
                    result = self._attach_live_reload(result)

                for func in self.after_request_funcs:
                    result = func(request, result) or result

                return self._process_outgoing(request, result, start_time)
            except Exception as e:
                if self.debug:
                    error_response = self._friendly_debug_error_response(e)
                    if self.debug:
                        self._set_no_cache_headers(error_response)
                    return self._process_outgoing(request, error_response, start_time)
                error_response = self._handle_error(request, 500, self._default_500)
                return self._process_outgoing(request, error_response, start_time)

        response = self._handle_error(request, 404, self._default_404)
        if self.debug:
            self._set_no_cache_headers(response)
        return self._process_outgoing(request, response, start_time)


    def __call__(self, environ, start_response):
        """
        WSGI application entry point
        :param environ:
        :param start_response:
        :return:
        """
        start_time = time.perf_counter()
        request = Request(environ)
        self._load_session(request)

        # Lightweight endpoint used by browser-side live reload polling.
        if request.path == '/__mallo_reload__':
            response = Response(
                self._reload_token,
                headers={
                    'Content-Type': 'text/plain',
                    'Cache-Control': 'no-store, no-cache, must-revalidate',
                    'Pragma': 'no-cache',
                    'Expires': '0'
                }
            )
            response = self._process_outgoing(request, response, start_time)
            return response.to_wsgi(start_response)

        for func in self.before_request_funcs:
            result = func(request)
            if isinstance(result, Response):
                if self.debug:
                    self._set_no_cache_headers(result)
                result = self._process_outgoing(request, result, start_time)
                return result.to_wsgi(start_response)

        # Match route before CSRF so route-level options can be applied.
        route_match = self.router.match(request.path, request.method, return_route=True)
        route_options = route_match[2].get('options', {}) if route_match else {}
        route_csrf = route_options.get('csrf')
        csrf_required = self.csrf_protect if route_csrf is None else bool(route_csrf)

        # CSRF protection for unsafe methods
        if not self._csrf_check(request, csrf_required=csrf_required):
            response = Response('<h1>403 Forbidden</h1>', status=403)
            if self.debug:
                self._set_no_cache_headers(response)
            response = self._process_outgoing(request, response, start_time)
            return response.to_wsgi(start_response)

        try:
            result = self._run_middleware_chain(
                request,
                lambda req: self._dispatch_request(req, start_time, route_match=route_match),
                self.middlewares
            )
            response = self._ensure_response(result)
            response = self._process_outgoing(request, response, start_time)
            return response.to_wsgi(start_response)
        except Exception as e:
            if self.debug:
                error_response = self._friendly_debug_error_response(e)
                if self.debug:
                    self._set_no_cache_headers(error_response)
                error_response = self._process_outgoing(request, error_response, start_time)
                return error_response.to_wsgi(start_response)
            error_response = self._handle_error(request, 500, self._default_500)
            error_response = self._process_outgoing(request, error_response, start_time)
            return error_response.to_wsgi(start_response)

    def _set_no_cache_headers(self, response):
        """
        Prevent browser caching while debugging.
        """
        response.headers['Cache-Control'] = 'no-store, no-cache, must-revalidate, max-age=0'
        response.headers['Pragma'] = 'no-cache'
        response.headers['Expires'] = '0'

    def _attach_live_reload(self, response):
        """
        Inject browser auto-refresh script into HTML responses in debug mode.
        """
        content_type = response.headers.get('Content-Type', '')
        if not isinstance(response.body, str):
            return response
        if 'text/html' not in content_type.lower():
            return response
        if '__mallo_reload__' in response.body:
            return response

        script = """
<script>
(function () {
  var token = null;
  function checkReload() {
    fetch('/__mallo_reload__?t=' + Date.now(), { cache: 'no-store' })
      .then(function (res) { return res.text(); })
      .then(function (value) {
        value = value.trim();
        if (token === null) {
          token = value;
          return;
        }
        if (token !== value) {
          window.location.reload();
        }
      })
      .catch(function () {});
  }
  setInterval(checkReload, 1000);
  checkReload();
})();
</script>
"""

        body = response.body
        if '</body>' in body:
            body = body.replace('</body>', script + '\n</body>', 1)
        else:
            body = body + script
        response.body = body
        return response


    def run(self, host='localhost', port=8000, debug=False, use_reloader=True):
        """
        Run the development server

        Args:
            host: Host to bind to
            port: Port to listen on
            debug: Enable debug mode
            use_reloader: Enable hot reload

        :param host:
        :param port:
        :param debug:
        :param use_reloader:
        :return:
        """
        self.debug = debug

        in_reloader_subprocess = os.environ.get('MALLO_HOT_RELOAD') == '1'
        if use_reloader and debug and not in_reloader_subprocess:
            # Start with hot reload
            self.hot_reloader = HotReloader(self)
            self.hot_reloader.run(host, port)
        else:
            # Start Normally
            print(f" * Running on http://{host}:{port}/")
            print(f" * Debug Mode: {'on' if debug else 'off'}")
            if debug:
                self._print_startup_diagnostics()

            class QuietReloadRequestHandler(WSGIRequestHandler):
                def log_message(self, format, *args):
                    try:
                        request_line = args[0] if args else ""
                        if "__mallo_reload__" in str(request_line):
                            return
                    except Exception:
                        pass
                    super().log_message(format, *args)

            server = make_server(host, port, self, handler_class=QuietReloadRequestHandler)
            try:
                server.serve_forever()
            except KeyboardInterrupt:
                print("\n * Server stopped")

    def _guess_mime_type(self, filename):
        """
        Guess MIME type from filename extension
        :param filename:
        :return:
        """
        ext_map ={
            '.html': 'text/html',
            '.css': 'text/css',
            '.js': 'application/javascript',
            '.json': 'application/json',
            '.png': 'image/png',
            '.jpg': 'image/jpeg',
            '.jpeg': 'image/jpeg',
            '.gif': 'image/gif',
            '.svg': 'image/svg+xml',
            '.txt': 'text/plain'
        }

        ext = os.path.splitext(filename)[1].lower()
        return ext_map.get(ext, 'application/octet-stream')



"""
Mallo - A lightweight Web framework with Hot reload
"""

from mallo.app import Mallo
from mallo.request import Request
from mallo.response import Response
from mallo.cli import cli
from mallo.template import render_template, render_template_file
from mallo.db import Database
from mallo.config import MalloConfig

__version__ = "0.1.0"
__all__ = ['Mallo', 'Request', 'Response', 'cli', 'render_template', 'render_template_file', 'Database', 'MalloConfig' ]



"""
Command line interface for Mallo framework.
"""

import argparse
import os
import re
import subprocess
import sys
import platform
from pathlib import Path
from textwrap import dedent


def _is_valid_project_name(name: str) -> bool:
    return bool(re.match(r"^[A-Za-z0-9._-]+$", name or ""))


def _write_file(path: Path, content: str, force: bool) -> None:
    if path.exists() and not force:
        raise FileExistsError(f"File already exists: {path}")
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(content, encoding="utf-8")


def create_project(name: str, template_folder: str = "templates", static_folder: str = "static", force: bool = False) -> int:
    """Create a new Mallo project scaffold safely."""
    if not name or not name.strip():
        print("Error: project name cannot be empty.")
        return 1
    if not _is_valid_project_name(name):
        print("Error: project name may only contain letters, numbers, '.', '_' or '-'.")
        return 1
    if not template_folder.strip() or not static_folder.strip():
        print("Error: template/static folder names cannot be empty.")
        return 1

    project_dir = Path(name).resolve()
    if project_dir.exists() and not project_dir.is_dir():
        print(f"Error: path exists and is not a directory: {project_dir}")
        return 1
    if project_dir.exists() and not force and any(project_dir.iterdir()):
        print(f"Error: directory is not empty: {project_dir}")
        print("Use --force to overwrite scaffold files.")
        return 1

    print(f"Creating new Mallo project: {name}")
    project_dir.mkdir(parents=True, exist_ok=True)

    app_content = dedent(
        f"""\
        \"\"\"
        Mallo application
        \"\"\"

        from mallo import Mallo

        app = Mallo(__name__, template_folder='{template_folder}', static_folder='{static_folder}')

        @app.get('/')
        def home(request):
            return app.render_template('index.html')

        @app.get('/hello/<name>')
        def hello(request, name):
            return f'<h1>Hello, {{name}}!</h1>'

        @app.get('/api/data')
        def api_data(request):
            return {{'message': 'Hello from Mallo!', 'status': 'success'}}

        if __name__ == '__main__':
            app.run(debug=True, use_reloader=True)
        """
    )

    index_content = dedent(
        """\
        <!DOCTYPE html>
        <html lang="en">
        <head>
          <meta charset="UTF-8" />
          <meta name="viewport" content="width=device-width, initial-scale=1" />
          <title>Mallo App</title>
          <link rel="stylesheet" href="/static/style.css" />
        </head>
        <body>
          <main class="wrap">
            <h1>Mallo is running</h1>
            <p>Your project scaffold is ready.</p>
            <p>Try <a href="/hello/World">/hello/World</a> and <a href="/api/data">/api/data</a>.</p>
          </main>
        </body>
        </html>
        """
    )

    css_content = dedent(
        """\
        :root {
          --bg: #f6f2ea;
          --surface: #fffdf8;
          --ink: #1a1510;
          --muted: #6b5f52;
          --line: #e8ddcd;
          --accent: #b6551d;
        }

        body {
          margin: 0;
          font-family: "Segoe UI", Tahoma, sans-serif;
          color: var(--ink);
          background: var(--bg);
        }

        .wrap {
          max-width: 700px;
          margin: 40px auto;
          padding: 20px;
          border: 1px solid var(--line);
          border-radius: 12px;
          background: var(--surface);
        }

        h1 { margin-top: 0; }
        p { color: var(--muted); }
        a { color: var(--accent); }
        """
    )

    readme_content = dedent(
        f"""\
        # {name}

        Generated by Mallo CLI.

        ## Run

        ```bash
        python app.py
        ```
        """
    )

    gitignore_content = dedent(
        """\
        __pycache__/
        *.pyc
        .venv/
        """
    )

    try:
        _write_file(project_dir / "app.py", app_content, force)
        _write_file(project_dir / template_folder / "index.html", index_content, force)
        _write_file(project_dir / static_folder / "style.css", css_content, force)
        _write_file(project_dir / "README.md", readme_content, force)
        _write_file(project_dir / ".gitignore", gitignore_content, force)
    except Exception as exc:
        print(f"Error: failed to create project files: {exc}")
        return 1

    print(f"Project '{name}' created successfully.")
    print(f"cd {name}")
    print("python app.py")
    return 0


def run_server(app_path: str, host: str, port: int, debug: bool, reload: bool) -> int:
    """Run a Mallo application via gunicorn."""
    if platform.system().lower().startswith("win"):
        print("Gunicorn is not supported on Windows. Use 'python app.py' or run on Linux/WSL.")
        return 1

    env = os.environ.copy()
    env["MALLO_DEBUG"] = "1" if debug else "0"
    env["MALLO_LIVE_RELOAD"] = "1" if debug else "0"

    cmd = [
        "gunicorn",
        app_path,
        "--bind",
        f"{host}:{port}",
    ]
    if reload:
        cmd.append("--reload")

    try:
        result = subprocess.run(cmd, env=env, check=False)
        return result.returncode
    except FileNotFoundError:
        print("Error: gunicorn is not installed. Install it with: pip install gunicorn")
        return 1


def cli() -> None:
    """Main CLI entry point."""
    parser = argparse.ArgumentParser(description="Mallo web framework CLI")
    subparsers = parser.add_subparsers(dest="command", help="Commands")

    create_parser = subparsers.add_parser("create", help="Create a new project")
    create_parser.add_argument("name", help="Project name")
    create_parser.add_argument("--template-folder", default="templates", help="Template folder name")
    create_parser.add_argument("--static-folder", default="static", help="Static folder name")
    create_parser.add_argument("--force", action="store_true", help="Overwrite scaffold files if they exist")

    run_parser = subparsers.add_parser("run", help="Run the application")
    run_parser.add_argument("app", nargs="?", default="app:app", help="WSGI app path, e.g. app:app")
    run_parser.add_argument("--host", default="localhost", help="Host to bind to")
    run_parser.add_argument("--port", type=int, default=8000, help="Port to listen on")
    run_parser.add_argument("--debug", action="store_true", default=True, help="Enable debug mode")
    run_parser.add_argument("--no-debug", action="store_true", help="Disable debug mode")
    run_parser.add_argument("--no-reload", action="store_true", help="Disable hot reload")

    args = parser.parse_args()

    if args.command == "create":
        code = create_project(
            args.name,
            template_folder=args.template_folder,
            static_folder=args.static_folder,
            force=args.force,
        )
        sys.exit(code)
    if args.command == "run":
        debug = args.debug and not args.no_debug
        reload = not args.no_reload and debug
        code = run_server(args.app, args.host, args.port, debug, reload)
        sys.exit(code)

    parser.print_help()


if __name__ == "__main__":
    cli()



"""
Setup script for Mallo framework
"""

from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = [line.strip() for line in fh if line.strip() and not line.startswith("#")]

setup(
    name="mallo",
    version="0.1.0",
    author="Akum betrand",
    author_email="betrandojong146@gmail.com",
    description="A lightweight web framework with hot reload",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Betrand-dev/mallo-fr.git",
    packages=find_packages(),
    package_data={
        "mallo": ["defaults/*.html"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
    ],
    python_requires=">=3.9",
    install_requires=requirements,
    entry_points={
        "console_scripts": [
            "mallo=mallo.cli:cli",
        ],
    },
    include_package_data=True,
    zip_safe=False,
)



"""
Simple template rendering engine for Mallo framework
"""
import os
import re
import html
from typing import Dict, Any

class TemplateEngine:
    """
    Simple template engine with variable substitution and basic control structures
    """

    def __init__(self):
        self.cache = {}

    def render(self, template_path: str, auto_reload: bool = False, auto_escape: bool = True, **context) -> str:
        """
        Render a template file with context variables
        Supports:
               {{ variable }} -variable substitution
               {% if conditions %}....{% endif %} - if statements
               {% for item in list %}...{% endfor %} - For loops

        :param template_path:
        :param context:
        :return:
        """
        if not os.path.exists(template_path):
            return _template_error_page(
                "Template not found",
                "Mallo could not find the template file.",
                template_path
            )

        current_mtime = os.path.getmtime(template_path)
        cached = self.cache.get(template_path)

        if cached and (not auto_reload or cached['mtime'] == current_mtime):
            template = cached['template']
        else:
            with open(template_path, 'r', encoding='utf-8') as f:
                template = f.read()
            self.cache[template_path] = {
                'template': template,
                'mtime': current_mtime
            }

        # Render the template
        try:
            return self._render_string(template, context, auto_escape=auto_escape)
        except Exception as exc:
            return _template_error_page(
                "Template render error",
                "Mallo failed while rendering this template.",
                f"{template_path}\n{type(exc).__name__}: {exc}"
            )

    def _render_string(self, template: str, context: Dict[str, Any], auto_escape: bool = True) -> str:
        """
        Render a template with context

        :param template:
        :param context:
        :return:
        """
        def resolve_var(var_name: str, ctx: Dict[str, Any]):
            parts = var_name.split('.')
            value = ctx
            for part in parts:
                if isinstance(value, dict):
                    value = value.get(part, '')
                else:
                    value = getattr(value, part, '')
            return value

        def render_value(expr: str, ctx: Dict[str, Any]) -> str:
            parts = [part.strip() for part in expr.split('|')]
            var_name = parts[0]
            filters = parts[1:]

            value = resolve_var(var_name, ctx)
            text = str(value)

            if 'safe' in filters:
                return text
            if auto_escape:
                return html.escape(text)
            return text

        token_re = re.compile(r'{%\s*(if|for|endif|endfor)\b([^%]*)%}')
        pos = 0
        tokens = []

        for match in token_re.finditer(template):
            if match.start() > pos:
                tokens.append(('text', template[pos:match.start()]))
            keyword = match.group(1)
            expr = match.group(2).strip()
            tokens.append((keyword, expr))
            pos = match.end()

        if pos < len(template):
            tokens.append(('text', template[pos:]))

        def parse_tokens(token_list, index=0):
            nodes = []
            while index < len(token_list):
                kind, value = token_list[index]
                if kind == 'text':
                    nodes.append(('text', value))
                    index += 1
                elif kind == 'if':
                    index += 1
                    inner, index = parse_tokens(token_list, index)
                    nodes.append(('if', value, inner))
                elif kind == 'for':
                    index += 1
                    inner, index = parse_tokens(token_list, index)
                    nodes.append(('for', value, inner))
                elif kind in ('endif', 'endfor'):
                    index += 1
                    return nodes, index
                else:
                    index += 1
            return nodes, index

        def eval_condition(expr: str, ctx: Dict[str, Any]) -> bool:
            expr = expr.strip()
            negate = False
            if expr.startswith('not '):
                negate = True
                expr = expr[4:].strip()
            value = resolve_var(expr, ctx)
            return (not value) if negate else bool(value)

        def render_nodes(nodes, ctx: Dict[str, Any]) -> str:
            output = []
            for node in nodes:
                kind = node[0]
                if kind == 'text':
                    output.append(re.sub(r'{{(.*?)}}', lambda m: render_value(m.group(1).strip(), ctx), node[1]))
                elif kind == 'if':
                    expr, inner = node[1], node[2]
                    if eval_condition(expr, ctx):
                        output.append(render_nodes(inner, ctx))
                elif kind == 'for':
                    expr, inner = node[1], node[2]
                    if ' in ' not in expr:
                        continue
                    var_name, list_name = [part.strip() for part in expr.split(' in ', 1)]
                    items = resolve_var(list_name, ctx)
                    if items is None:
                        continue
                    for item in items:
                        new_ctx = dict(ctx)
                        new_ctx[var_name] = item
                        output.append(render_nodes(inner, new_ctx))
            return ''.join(output)

        nodes, _ = parse_tokens(tokens)
        return render_nodes(nodes, context)

# Global template engine instance
_engine = TemplateEngine()

def _template_error_page(title: str, message: str, detail: str = "") -> str:
    safe_title = html.escape(title)
    safe_message = html.escape(message)
    safe_detail = html.escape(detail)
    return f"""
<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="UTF-8" />
  <meta name="viewport" content="width=device-width, initial-scale=1" />
  <title>Mallo Template Error</title>
  <style>
    body {{ margin: 0; font-family: Segoe UI, Arial, sans-serif; background: #f6f3ee; color: #1f1a15; }}
    .wrap {{ max-width: 760px; margin: 28px auto; padding: 0 16px; }}
    .card {{ background: #fffdf9; border: 1px solid #e8dece; border-radius: 14px; padding: 18px; }}
    .tag {{ display: inline-block; border-radius: 999px; padding: 4px 10px; background: #f8e8de; color: #b6521a; font-size: 12px; font-weight: 700; letter-spacing: .05em; text-transform: uppercase; }}
    h1 {{ margin: 10px 0 8px; font-size: 26px; }}
    p {{ margin: 0 0 10px; color: #5f5549; }}
    pre {{ margin: 0; background: #1a1f29; color: #eaf0ff; border-radius: 10px; padding: 12px; white-space: pre-wrap; word-break: break-word; }}
  </style>
</head>
<body>
  <div class="wrap">
    <div class="card">
      <span class="tag">Template Error</span>
      <h1>{safe_title}</h1>
      <p>{safe_message}</p>
      <pre>{safe_detail}</pre>
    </div>
  </div>
</body>
</html>
"""

def render_template(
    template_name: str,
    template_folder: str = 'templates',
    auto_reload: bool = False,
    auto_escape: bool = True,
    **context
) -> str:
    """
    Render a template from the configured template folder.

    Args:
        template_name: Template name relative to template_folder (e.g. "index.html")
        template_folder: Folder that contains templates (default: "templates")
        auto_reload: Reload template file when changed
        auto_escape: Escape variables by default
        **context: Context variables for rendering
    """
    if not os.path.isdir(template_folder):
        return _template_error_page(
            "Template folder not found",
            "Mallo expected a template folder but it does not exist.",
            template_folder
        )

    template_path = os.path.join(template_folder, template_name)
    return render_template_file(
        template_path,
        auto_reload=auto_reload,
        auto_escape=auto_escape,
        **context
    )

def render_template_file(template_path: str, auto_reload: bool = False, auto_escape: bool = True, **context)->str:
    """
    Render a template from file path with context

    :param template_path:
    :param context:
    :return:
    """
    return _engine.render(template_path, auto_reload=auto_reload, auto_escape=auto_escape, **context)



from mallo.router import Router


def test_url_for():
    router = Router()

    def user(request, id):
        return f'User {id}'

    router.add_route('/user/<int:id>', user, ['GET'])
    assert router.url_for('user', id=5) == '/user/5'
